const http = require('http')
const { Semantics, Semantic } = require('./src/semantics')
const { Generators, Generator } = require('./src/generators')
const Digraph = require('./src/digraph')
const fetch = require('node-fetch')
const base64 = require('base-64')
const fs = require('fs')
const { ArgumentParser } = require('argparse')
const deepEqual = require('deep-equal')
const _ = require('lodash')
const Lines = require('./lines')

const fixJSON = (json) => JSON.parse(JSON.stringify(json))
const compareJSON = (j1, j2) => JSON.stringify(fixJSON(j1)) == JSON.stringify(fixJSON(j2))

const matching = (actual, expected) => {
  if (!deepEqual(fixJSON(actual), fixJSON(expected))) {
    return false
  }
  return true
}

const analyzeMetaData = (right, wrong) => {
  rpriorities = right.priorities
  wpriorities = wrong.priorities
  npriorities = []
  n = Math.min(rpriorities.length, wpriorities.length)
  for (let i = 0; i < n; ++i) {
    rp = rpriorities[i]
    wp = wpriorities[i]
    if (_.isEqual(rp[0], wp[0])) {
      continue
    }

    rop = rp[0]
    rchoices = rp[1]
    wop = wp[0]
    wchoices = wp[1]
    if (wchoices.find((wc) => _.isEqual(wc, rop))) {
      _.remove(wchoices, (wc) => _.isEqual(wc, rop))
      wchoices.push(rop)
      npriorities.push(wchoices)
      break
    }
  }
  return npriorities
}

const processContexts = (contexts, params) => {
  contextsPrime = []
  generateds = []
  for (context of contexts) {
    const result = processContext(context, params)
    contextsPrime.push(result.context)
    generateds.push(result.generated)
  }
  return { contexts: contextsPrime, generated: generateds }
}

const processContext = (context, { objects = {}, generators = [], semantics = [], config }) => {
  // map to hash
  config = config || {}
  if (config.config) {
    config = config
  }
  const response = {} // NA but passed in
  generators = new Generators(generators.map((g) => new Generator(g[0], g[1], g[2])))
  semantics = new Semantics(semantics.map((g) => new Semantic(g[0], g[1], g[2])))
  const hierarchy = new Digraph((config.config || {}).hierarchy || [])

  context = semantics.apply(objects, context, hierarchy, config, response)
  const generated = generators.apply(objects, context, hierarchy, config, response)[0]
  const paraphrases = generators.apply(objects, context, hierarchy, config, response, { paraphrase: true, response: false })[0]
  const responses = []
  if (context.response) {
    responses = generated
  }
  return { context, generated, paraphrases, responses }
}

const convertToStable = (objects) => {
  stableObjects = Object.assign({}, objects)
  stableObjects.namespaced = {}
  counter = 0
  Object.keys(objects.namespaced).forEach((uuid) => {
    stableObjects.namespaced[`${counter}`] = objects.namespaced[uuid]
    counter += 1
  })
  return stableObjects
}

const writeTest = (fn, query, objects, generated, paraphrases, responses, contexts, metadata) => {
  let tests = {}
  if (fs.existsSync(fn)) {
    tests = JSON.parse(fs.readFileSync(fn))
  }
  tests[query] = { generated, paraphrases, responses, contexts, objects: convertToStable(objects), metadata }
  fs.writeFileSync(fn, JSON.stringify(tests, null, 2), { encoding: 'utf8', flag: 'w+' })
}

const _process = (config, query, { writeTests, testsFN, errorHandler = defaultErrorHandler } = {}) => {
  const url = config._server
  const key = config._key
  retries = 2
  writeTests = writeTests || false

  // config.initialize({force: false})
  // config.initialize({force: false})
  // ensure same start state
  if (writeTests) {
    config.rebuild()
  }
  const data = Object.assign({ key, version: '3' }, config.config)
  data.utterance = query
  const generators = new Generators((data.generators || []).map((g) => new Generator(g[0], g[1], g[2])))
  delete data.generators
  const semantics = new Semantics((data.semantics || []).map((g) => new Semantic(g[0], g[1], g[2])))
  delete data.semantics

  const doWithRetries = (n) => {
    return fetch(`${url}/process`, {
      method: 'POST',
      body: JSON.stringify(data),
      timeout: 1000 * 60 * 5, // it does not respect this timeout so that's why I have the retries
      headers: {
        mode: 'no-cors',
        'Content-Type': 'application/json'
      }
    }).then(result => {
      if (result.ok) {
        return result.json().then((json) => {
          return JSON.parse(JSON.stringify(json))
        })
      }
      if (result.status == 504) {
        if (n == 0) {
          throw `Error ${result.status} - ${result.statusText}`
        } else {
          return doWithRetries(n - 1)
        }
      }
      if (result.status >= 500 && result.status < 600) {
        throw `Error ${result.status} - ${result.statusText}. If you are using the researcher version try restarting your entodicton server. If takes a few minutes to be up.`
      } else {
        throw `Error ${result.status} - ${result.statusText}`
      }
    })
  }

  const todo = doWithRetries(retries)
  const hierarchy = new Digraph((config.config || {}).hierarchy || [])
  return todo
    .then(json => {
      json.contexts = json.results
      delete json.results
      return json
    })
    .then(json => {
      if (json.status != 200) {
        //throw json
        //const error = { query, error: [json] }
        //errorHandler(json)
        throw json;
      } else if (json.contexts) {
        // TODO fix this name to contextsPrime
        const contextsPrime = []
        const generatedPrime = []
        const paraphrasesPrime = []
        const responsesPrime = []
        const contexts = json.contexts
        for (let j = 0; j < contexts.length; ++j) {
          const context = contexts[j]
          let contextPrime = context
          context.topLevel = true
          try {
            if (!config.get('skipSemantics')) {
              contextPrime = semantics.apply(config.get('objects'), context, hierarchy, config, json)
            }
            const generated = generators.apply(config.get('objects'), contextPrime, hierarchy, config, json)[0]
            const paraphrases = generators.apply(config.get('objects'), contextPrime, hierarchy, config, json, { paraphrase: true, response: false })[0]
            contextsPrime.push(contextPrime)
            generatedPrime.push(generated)
            paraphrasesPrime.push(paraphrases)
            if (contextPrime.response) {
              responsesPrime.push(generated)
            } else {
              responsesPrime.push('')
            }

            // add results to processed list
            config.config.objects.processed = config.config.objects.processed || []
            config.config.objects.processed = config.config.objects.processed.slice(0, 5)
            config.config.objects.processed.unshift({ context: contextPrime, paraphrases: paraphrases, responses: responsesPrime })        
          } catch (e) {
            e.context = contextPrime
            e.logs = json.logs
            e.metadata = json.metadata
            if (json.trace) {
              e.trace = json.trace
            }
            throw e
          }
        }

        json.contexts = contextsPrime
        json.generated = generatedPrime
        json.paraphrases = paraphrasesPrime
        json.responses = responsesPrime
        generators.logs.forEach((log) => json.logs.push(log))
        semantics.logs.forEach((log) => json.logs.push(log))

        if (writeTests) {
          writeTest(testsFN, query, config.config.objects, json.generated, json.paraphrases, json.responses, json.contexts, json.metadata)
        }
      }
      /*
      if (json.contexts) {
        jsons = json.results[0]
        json.generated = json.generated[0]
      }
      */
      return json
    }).catch((error) => {
      // throw `While processing "${query}" got error ${JSON.stringify(error, null, 2)}`
      /*
      debugger;
      error.query = query;
      throw error;
      */
      error.query = query
      //errorHandler(error)
      throw error
    })
}

const runTest = (config, test, expected) => {
  // initialize in between test so state is not preserved since the test was adding without state
  config.rebuild()

  errorHandler = (error) => {
    if (error.metadata) {
      const priorities = analyzeMetaData(expected.metadata, error.metadata)
      const log = `Hint, if the results are flakey try adding the specified priorities ${JSON.stringify(priorities)}`
      error.logs.push(log)
    }
    defaultErrorHandler(error)
  }

  return _process(config, test, { errorHandler }).then((result) => {
    const okay = true
    const expected_objects = convertToStable(expected.objects)
    const actual_objects = convertToStable(config.config.objects)
    let failed =
            !matching(result.generated, expected.generated) ||
            !matching(result.paraphrases, expected.paraphrases) ||
            !matching(result.responses, expected.responses) ||
            !matching(result.contexts, expected.contexts) ||
            !matching(actual_objects, expected_objects)
    if (!failed) {
      if (config._afterTest) {
        failed = config._afterTest({ query: test, expected, actual: result, config })
        if (failed) {
          return {
            utterance: test,
            errorFromAfterTest: failed,
            expected: { generated: expected.generated, results: expected.contexts, objects: expected_objects },
            actual: { generated: result.generated, results: result.contexts, objects: actual_objects }
          }
        }
      }
    }

    let analysis
    if (expected.metadata && result.metadata && failed) {
      const priorities = analyzeMetaData(expected.metadata, result.metadata)
      const log = `Hint, if the results are flakey try adding the specified priorities ${JSON.stringify(priorities)}`
      result.logs.push(log)
    }
    if (failed) {
      return {
        utterance: test,
        expected: { generated: expected.generated, results: expected.contexts, objects: expected_objects },
        actual: { generated: result.generated, results: result.contexts, objects: actual_objects }
      }
    }
  }).catch((error) => {
    if (error.metadata) {
      const priorities = analyzeMetaData(expected.metadata, error.metadata)
      const log = `Hint, if the results are flakey try adding the specified priorities ${JSON.stringify(priorities)}`
      error.logs.push(log)
    }
    throw error
  })
}

const runTestsHelper = (config, tests, todo, failed, { verbose, stopAtFirstError}) => {
  if (todo.length == 0) {
    return failed
  }
  const test = todo.pop()

  if (verbose) {
    console.log(test)
  }

  return runTest(config, test, tests[test]).then((result) => {
    if (result != null) {
      failed.push(result)
      if (stopAtFirstError) {
        return failed
      }
    }
    return runTestsHelper(config, tests, todo, failed, { verbose, stopAtFirstError })
  })
}

const runTests = (config, testFile, juicyBits) => {
  const tests = JSON.parse(fs.readFileSync(testFile))
  const { beforeTests, afterTests } = juicyBits;
  beforeTests()
  return runTestsHelper(config, tests, Object.keys(tests), [], juicyBits).then( (v) => {
    afterTests()
    return v
  })
}

const saveTest = (fn, config, test, expected, beforeQuery) => {
  config.rebuild()
  beforeQuery( { query: test, isModule: false } )
  return _process(config, test).then((result) => {
    const okay = true
    const expected_objects = expected.objects
    const actual_objects = config.config.objects
    writeTest(fn, test, config.config.objects, result.generated, result.paraphrases, result.responses, result.contexts, result.metadata)
  })
}
const saveTestsHelper = (fn, config, tests, todo, beforeQuery) => {
  if (todo.length == 0) {
    return
  }
  const test = todo.pop()
  config.rebuild()
  return saveTest(fn, config, test, tests[test], saveQuery).then((result) => {
    // initialize in between test so state is not preserved since the test was adding without state
    // config.initialize({force: true})
    config.rebuild()
    return saveTestsHelper(fn, config, tests, todo, beforeQuery)
  })
}

const saveTests = (config, testFile, beforeQuery) => {
  const tests = JSON.parse(fs.readFileSync(testFile))
  return saveTestsHelper(testFile, config, tests, Object.keys(tests), [], beforeQuery)
}

const showExamples = (testFile) => {
  const tests = JSON.parse(fs.readFileSync(testFile))
  Object.keys(tests).forEach((test) => console.log(test))
}

const showInfo = (name, description, config, testFile) => {
  const tests = JSON.parse(fs.readFileSync(testFile))
  const includes = config.configs.slice(1).map((km) => km.config.name)
  const examples = Object.keys(tests)
  console.log(JSON.stringify({ name, description, examples, includes, demo: config.demo }, null, 2))
}

const submitBugToAPI = (subscription_id, subscription_password, config) => {
  console.log('********* Submitting bug *********')
  const body = { description: config.config.description, config: config.config }
  // fetch('http://localhost:5000/bug', {
  fetch('https://thinktelligence.com/api/bug', {
    method: 'POST',
    body: JSON.stringify(body),
    headers: {
      mode: 'no-cors', // Type of mode of the request
      'Content-Type': 'application/json', // request content type
      Authorization: 'Basic ' + base64.encode(subscription_id + ':' + subscription_password)
    }
  }).then((result) => result.json())
    .then((json) => {
      if (json.status == 404) {
        console.log(`Error submitting the bug: ${json.status} ${json.statusText}`)
      } else {
        console.log(`New bug id id ${json.id}`)
      }
    })
}

const submitBug = (subscription_id, subscription_password, config, utterance, retries = 2) => {
  // TODO remove these from the config
  const properties = ['expected_contexts', 'expected_generated']
  properties.forEach((property) => {
    if (!config.get('expected_contexts')) {
      throw 'Missing property expected_contexts'
    }
  })

  return _process(config, utterance)
    .then((responses) => {
      let hasError = false
      if (!matching(responses.contexts, config.config.expected_contexts)) {
        console.log('JSON does not match')
        console.log('actual', JSON.stringify(responses.contexts))
        console.log('expected', JSON.stringify(config.config.expected_contexts))
        hasError = true
      }
      if (!matching(responses.generated, config.config.expected_generated)) {
        console.log('Generated does not match')
        console.log('actual', JSON.stringify(responses.generated))
        console.log('expected', JSON.stringify(config.config.expected_generated))
        hasError = true
      }

      if (hasError) {
        submitBugToAPI(subscription_id, subscription_password, config)
      } else {
        throw '\n\n *** The expected contexts matched the generated contexts so the bug was not submitted since its working. ***\n\n:'
      }
    })
    .catch((error) => {
      console.log('Error', error)
      throw error
    })
}

const isEntryPoint = () => {
  return require.main === module
}

const defaultErrorHandler = (error) => {
  if (error.query) {
    console.log('query: ', error.query)
  }

  if (error.logs) {
    console.log('\nlogs: ')
    for (const log of error.logs) {
      console.log('\n  ', log)
    }
  }

  if (error.trace) {
    console.log('trace: ', error.trace)
  }

  if (error.stack) {
    console.log('error: ')
    console.log(error.stack)
  }

  if (error.error) {
    console.log('error: ')
    for (const e of error.error) {
      console.log('\n  ', e)
    }
  }


  //throw error
}

const defaultProcess = ({ config, errorHandler }) => (promise) => {
  return promise
    .then((responses) => {
      if (responses.errors) {
        console.log('Errors')
        responses.errors.forEach((error) => console.log(`    ${error}`))
      }
      console.log('This is the global objects from running semantics:\n', config.objects)
      if (responses.logs) {
        console.log('Logs')
        responses.logs.forEach((log) => console.log(`    ${log}`))
      }
      console.log(responses.trace)
      console.log('objects', JSON.stringify(config.get('objects'), null, 2))
      console.log('--- The contexts are ----------')
      console.log(JSON.stringify(responses.contexts, null, 2))
      console.log('')
      /*
      console.log('--- The generated strings are ----------')
      for (const response of responses.generated) {
        console.log(response)
      }
      */
      console.log('')
      const widths = [60, 10, 72]
      const lines = new Lines(widths)
      lines.setElement(0, 0, '--- The paraphrases are ----------')
      lines.setElement(0, 2, '--- The response strings are ----------')
      lines.log()
      for (let i = 0; i < responses.paraphrases.length; ++i) {
        lines.setElement(0, 0, responses.paraphrases[i])
        if ((responses.responses[i] || []).length > 0) {
          lines.setElement(0, 2, responses.responses[i])
        } else {
          lines.setElement(0, 2, '')
        }
        lines.log()
      }
    })
    .catch((error) => {
      defaultErrorHandler(error)
    })
}

const knowledgeModule = ({ 
    module: moduleFromJSFile, 
    name, 
    description, 
    config, 
    demo, 
    test, 
    errorHandler = defaultErrorHandler, 
    process: processResults = defaultProcess, 
    stopAtFirstFailure = true, 
    beforeQuery = () => {}, 
    beforeTests = () => {},
    afterTests = () => {},
    beforeTest = () => {}, 
    afterTest = () => {},
  } = {}) => {
  const isProcess = require.main === moduleFromJSFile

  if (!moduleFromJSFile) {
    throw "'module' is a required parameter. The value should be either 'module' or a lambda that will be called when the file is acting as a module."
  }
  if (!config) {
    throw "'config' is a required parameter. The value should the config that defines the knowledge module."
  }
  if (!name) {
    throw "'name' is a required parameter. The value should the name of the knowledge module."
  }
  if (!description) {
    throw "'description' is a required parameter. The value should the description of the knowledge module."
  }
  if (!test) {
    throw "'test' is a required parameter. The value should the path to the file used to store the tests of the knowledge module."
  }

  let module
  if (_.isFunction(moduleFromJSFile)) {
    module = moduleFromJSFile
  } else {
    module = () => {
      config.rebuild({ isModule: true })
      moduleFromJSFile.exports = config
    }
  }
  processResults = processResults({ config, errorHandler })
  config.name = name
  config.description = description
  config.demo = demo
  if (isProcess) {
    if (fs.existsSync(test)) {
      config.tests = JSON.parse(fs.readFileSync(test))
    } else {
      config.tests = {}
    }
    // setup();
    const parser = new ArgumentParser({
      description: 'Entodicton knowledge module'
    })

    parser.add_argument('-t', '--test', { action: 'store_true', help: 'Run the tests. Create tests by running with the --query + --save flag' })
    parser.add_argument('-tv', '--testVerbose', { action: 'store_true', help: 'Run the tests in verbose mode. Create tests by running with the --query or --loop + --save flag' })
    parser.add_argument('-n', '--count', { help: 'Number of times to run the tests. Default is one. Use this to check for flakey test. If possible the system will print out a message with the word "hint" suggesting how to fix the problem' })
    parser.add_argument('-l', '--loop', { action: 'store_true', help: 'Run a loop so that multiply queries may be run' })
    parser.add_argument('-i', '--info', { action: 'store_true', help: 'Print meta-data for the module' })
    parser.add_argument('-v', '--vimdiff', { action: 'store_true', help: 'For failures run vimdiff' })
    parser.add_argument('-g', '--greg', { action: 'store_true', help: 'Set the server to be localhost so I can debug stuff' })
    parser.add_argument('-r', '--retrain', { action: 'store_true', help: 'Get the server to retrain the neural nets' })
    parser.add_argument('-q', '--query', { help: 'Run the specified query' })
    parser.add_argument('-s', '--save', { action: 'store_true', help: 'When running with the --query flag this will save the current run to the test file. When running without the --query flag all tests will be run and resaved.' })
    parser.add_argument('-d', '--debug', { action: 'store_true', help: 'When running with the --debug flag this set the debug flag in the config' })

    args = parser.parse_args()
    args.count = args.count || 1

    if (args.greg) {
      config.server('http://localhost:3000', '6804954f-e56d-471f-bbb8-08e3c54d9321')
    }

    if (args.debug) {
      config.config.debug = true
    }

    if (args.retrain) {
      config.config.retrain = true
    }

    if (!args.query && !args.test && !args.info && args.save) {
      saveTests(config, test, beforeQuery)
    } else if (args.info) {
      showInfo(name, description, config, test)
    } else if (args.test || args.testVerbose) {
      // TODO make test always a string
      if (typeof test === 'string') {
        const l = (n) => {
          if (n == 0) {
            return
          }
          runTests(config, test, { verbose: args.testVerbose, stopAtFirstError: true, beforeQuery, beforeTests, afterTests, beforeTest, afterTest }).then((failures) => {
            if (args.vimdiff) {
              const path = '.'
              const actual = failures.map((one) => one.actual)
              fs.writeFileSync(`${path}/actual.json`, JSON.stringify(actual, 0, 2))
              const expected = failures.map((one) => one.expected)
              fs.writeFileSync(`${path}/expected.json`, JSON.stringify(expected, 0, 2))
              console.log(`vimdiff ${path}/actual.json ${path}/expected.json`)
              /*
              exec(`vimdiff ${path}/actual.json ${path}/expected.json`) ...
                if (error) {
                  console.log(`error: ${error.message}`);
                  return;
                }
                if (stderr) {
                  console.log(`stderr: ${stderr}`);
                  return;
                }
                console.log(`stdout: ${stdout}`);
                });
              */
            }
            const contexts = { failures: failures }
            console.log(JSON.stringify(contexts))
            l(n - 1)
          })
        }
        l(args.count)
      } else {
        test()
      }
    } else if (args.loop) {
      const readline = require('readline').createInterface({ input: process.stdin, output: process.stdout })
      const f = () => readline.question('Enter query? (newline to quit) ', query => {
        query = query.trim()
        if (query.length == 0) {
          return readline.close()
        }
        //const promise = processResults(_process(config, query, { testsFN: test }))
        const promise = _process(config, query, { testsFN: test }).then( (results) => {
          console.log(results.responses.join(' '))
        })
        if (!('then' in promise)) {
          throw 'Return a promise from process in the definition of knowledgeModule'
        }
        promise.then(() => f())
      })
      f()
    } else if (args.query) {
      beforeQuery( { query: args.query, isModule: false } )
      processResults(_process(config, args.query, { writeTests: args.save, testsFN: test }))
    }
  } else {
    module()
  }
}

module.exports = {
  process: _process,
  submitBug,
  processContext,
  processContexts,
  runTests,
  knowledgeModule,
  Semantics,
  Semantic,
  Generators,
  Generator,
  Digraph,
  analyzeMetaData
}
